
/* Copyright (C) 1998 IrseeSoft

  This file is part of GNU Ghostscript.

  GNU Ghostscript is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
  anyone for the consequences of using it or for whether it serves any
  particular purpose or works at all, unless he says so in writing.  Refer to
  the GNU General Public License for full details.

  Everyone is granted permission to copy, modify and redistribute GNU
  Ghostscript, but only under the conditions described in the GNU General
  Public License.  A copy of this license is supposed to have been given to
  you along with GNU Ghostscript so you can know your rights and
  responsibilities.  It should be in a file named COPYING.  Among other
  things, the copyright notice and this notice must be preserved on all
  copies.

*/
/*
 * gdevturboprint.c
 *
 *  Florian Zeiler (C) 1998 IrseeSoft
 *
 *  Turboprint 7 printer driver - this is Amiga platform specific
 *
 *  This device prints via the Amiga printer driver system "Turboprint" by IrseeSoft.
 *  Turboprint version 7 or higher is required. This driver does not print
 *  via the Workbench printer.device - use amiga_printer for this purpose.
 *
 *  There is only one device named "turboprint" for B/W and color modes.
 *
 *  Command line parameters - defined on the command line e.g. -dBitsPerPixel=24
 *
 *  -dBitsPerPixel=1   monochrome printing, grayscales generated by ghostscript's
 *                     halftoning - fastest printing method.
 *                 8   grayscale printing using Turboprint's dithering
 *                     or color printing combining ghostscript's halftoning and
 *                     Turboprint's dithering - uses less RAM
 *                 16  color printing using Turboprint's dithering; uses less RAM
 *                     than 24 bit but lower quality
 *                 24  color printing using Turboprint's dithering; highest quality;
 *                     this is the default
 *
 *  -dColor=0   grayscale printing
 *          1   colour printing, default
 *
 *  -rDPI       sets the x and y resolution to DPI; if resolution is not available
 *              the next lower resolution is used
 *  -rXDPIxYDPI sets the x resolution to XDPI and the y resolution to YDPI
 *
 *  -dDensity=1..7 Selects the print resolution according to the density setting in
 *                 Turboprefs
 *  -dQuality=0..9 Selects the print quality according to the setting in Turboprefs
 *
 *  -dZoomX=1..999 Selects a zoom factor
 *  -dZoomY=1..999
 *
 *  -dMoveX=... move the printout on the page - values are in 1/1000 Inch
 *		negative values: nove left / up
 *  -dMoveY=... positive values: move printout right / down
 */

/*
whoosh777@blueyonder.co.uk, this is a major rejuvenation of Florian Zeiler's 
original work for Turboprint printing via AFPL Ghostscript version 8 upwards.

Avoid using this for GNU Ghostscript unless it functions AS-IS, (GNU GS is a 
bit pointless if you have AFPL, AFPL GS is where the GS action is),

trying to change this for GNU GS could cause a lot of problems,

Dont use this for earlier AFPL GS's because it may well fail, GS8 has 
changed a lot internally from earlier GS's, though its not too clear 
when the major changes occurred,

The original code has been retained as much as possible. 

So this is a fixing of the malfunctions rather than trying to rewrite it,

When I begun the device wouldnt print a single pixel when compiled AS-IS 
into GS8. This was due to major changes in the internal mechanism of GS 
since the original -sDEVICE=turboprint was done. Also the original code 
took a lot of liberties which were fine for GS510, but for GS8 they were 
totally wrong, so totally wrong that 0 pixels got printed.

It now prints flawlessly in 24 bit truecolour and is currently in use by 
a lot of people.

The work was achieved by a lot of experimenting and feedback done long 
range for me by Michael Merkel and Peer Richter,

The rejuvenation is via a lot of quite Painful technical changes,
that being Painful with a capital P!
the original code was making a lot of assumptions that no longer apply,

Changes to what it says in the above comment:

these next statements override what Florian's original comment says:

The device will now print via Turboprint 3,4,5,6,7 the original device only 
printed via Turboprint 7. Careful study of the code enabled me to get the 
device to reach back into the depths of time as far as version 3.

The device is now split up into 3 devices: 
-sDEVICE=tp1  => 1 bit b/w , very fast, very good for b/w text, not for photos!
-sDEVICE=tp8  => 8 bit gray, for gray photorealism, 
-sDEVICE=tp24 => 24 bit truecolour, maximum quality from your printer,

splitting it into 3 will ensure this device will run correctly for many 
years of new versions of GS.

Having 3 devices is in line with other GS devices such as png and jpeg,
and I use the same approach for my whoosh1, whoosh8 and whoosh24 viewers,

most people will only be using tp24 which is the default + best quality option,

2 parameters have been removed: -dBitsPerPixel and -dColor,
they are now part of a complicated internal fix,
these parameters were quite confusing and the identical effect is 
now achieved by selecting one of tp1, tp8, tp24 ie 6 decisions have been 
replaced by 3,

16 bit printing has been removed as its totally pointless, 
memory overhead is not an issue for printing due to "banding" 
so for a small overhead you move from 65536 colours to 16777216 colours,

16 bit is also irregular because green gets 6 bits vs 5 bits for red + blue,
15 bit is  regular but gives a mere 32768 colours, so is also pointless

many decisions have been made and user feedback tells me they are 
the right decisions.

the code is currently very messy but its correct!

some of the commented out stuff could get brought back in so 
avoid removing seemingly irrelevant stuff,

at some point I will tidy it up, but I have to be very careful I 
dont damage anything,
*/

#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/asl.h>

/* apr19, for error messages, Mr. Zeillers debug things 
#define DEBUG
*/
/* for deb() function */
/* activates abandoned stuff,
#define XTURBO
*/
/* apr19 whoosh's debug things,
#define DEB
*/

#ifdef DEB
static void myputstr(char *mess )
{
PutStr(mess);Flush(Output());
}
#else
#define myputstr(mess)
#endif

#include "gdevprn.h"
#include "gdevpccm.h"	/* COLOR256 colormap support */

/* Amiga specific includes */

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <graphics/gfx.h>
#include <graphics/view.h>
#include <graphics/gfxmacros.h>
#include <graphics/gfxbase.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/preferences.h>
#include <devices/printer.h>
#include <devices/prtbase.h>

/* whoosh amigafeb2003: replaced these with the next lot
#include <clib/intuition_protos.h>
#include <clib/layers_protos.h>
#include <clib/gadtools_protos.h>
#include <clib/exec_protos.h>
#include <clib/locale_protos.h>
#include <clib/graphics_protos.h>
#include <clib/dos_protos.h>
*/

#include "../whoosh/protos.h"
/*
2005
#include <proto/intuition.h>
#include <proto/layers.h>
#include <proto/gadtools.h>
#include <proto/exec.h>
#include <proto/locale.h>
#include <proto/graphics.h>
#include <proto/dos.h>
*/
#include <stdio.h>
#include <string.h>


/* whoosh function: default is a keyword! */
static int env_int( char *name , int defaultval )
{
int ret ;
unsigned char *c_env ;

c_env = AllocVec( 32 , MEMF_PUBLIC | MEMF_CLEAR ) ;

/* dont use calloc, it is verboten! */

if( GetVar( name , c_env , 30 , 0 ) > 0 )ret = atoi( c_env ) ;
else ret = defaultval ;

FreeVec( c_env ) ;

return( ret ) ;
}

static int env_exists( char *name )
{
unsigned char *c ; int ret ;

c = AllocVec( 32 , MEMF_PUBLIC | MEMF_CLEAR ) ;

if( GetVar( name , c , 30 , 0 ) > 0 )ret = 1 ;
else ret = 0 ;

FreeVec( c ) ;

return( ret ) ;
}

/* 
tp??x : dont remove or change!
*/
static int turbostring( char *mess )
{
if( 0==strcmp( mess , "tp16x" ) )return( 1 ) ;
if( 0==strcmp( mess , "tp24x" ) )return( 1 ) ;
return( 0 ) ;
}

/*
#define		WHOOSHES	"tp_other"
*/
/* whoosh: */
#define TP_ERRORS "tp_errors" 

#include <whoosh_deb.ch>
#include "../whoosh/turboprint.h"

extern struct ExecBase *SysBase;

//#define DEBUG 1
#ifdef DEBUG
#define pdebug(string) tp_faultmessage((gx_device_printer *)pdev, string)
#define adrdebug(string, adr) tp_printadr((gx_device_printer *)pdev, string, adr)
#define intdebug(string, adr) tp_printint((gx_device_printer *)pdev, string, (adr))
#define inchdebug(string, adr) 	\
tp_printdouble((gx_device_printer *)pdev, string, ((adr) * 2.54) , "cm") ;

#else
#define pdebug(string)
#define adrdebug(string, adr)
#define intdebug(string, adr)
#define inchdebug(string,adr)
#endif

static dev_proc_open_device(turboprint_open);
static dev_proc_close_device(turboprint_close);
static dev_proc_print_page(turboprint_print_page);
static dev_proc_get_params(turboprint_get_params);
static dev_proc_put_params(turboprint_put_params);
static dev_proc_map_rgb_color(turboprint_24_map_rgb_color);
static dev_proc_map_rgb_color(turboprint_16_map_rgb_color);
static dev_proc_map_color_rgb(turboprint_map_color_rgb);
static dev_proc_print_page_copies(turboprint_print_page_copies);

/* whoosh: this can be deleted but I have left it in */
static gx_device_procs prn_turboprint_procs = {
        turboprint_open,
        gx_default_get_initial_matrix,
        gx_default_sync_output,
        gdev_prn_output_page,
        turboprint_close,
        NULL,	/* map_rgb_color */
     	NULL,	/* map_color_rgb */
        NULL,   /* fill_rectangle */
        NULL,   /* tile_rectangle */
        NULL,   /* copy_mono */
        NULL,   /* copy_color */
        NULL,   /* draw_line */
        gx_default_get_bits,
	turboprint_get_params,
        turboprint_put_params
};

static int tp_openprinterdevice(gx_device_printer *pdev);
static int tp_closeprinterdevice(gx_device_printer *pdev);
static int tp_setprefs(gx_device_printer *pdev);
static int tp_getres(gx_device_printer *pdev);
static int tp_noprintreq(gx_device_printer *pdev);
static int tp_faultmessage(gx_device_printer *pdev, char *string);
static int tp_printadr(gx_device_printer *pdev, char *string, char *adr);
static int tp_printint(gx_device_printer *pdev, char *string, int i);
static int
tp_printdouble(gx_device_printer *pdev, char *string, double i , char *string2 ) ;

static int tp_loadturboprint(gx_device_printer *pdev);
static int tp_build256colormap(gx_device_printer *pdev, struct ColorMap *colormap);
static int tp_trim_bitmap(gx_device_printer *pdev, byte *buffer, int bpr, int width, int height, int *maxheight);

/* here we define our device specific data structures and variables */
/* all variables are stored in the gx_device_turboprint structure */

/****** Turboprint **********/

#define MYMATCHWORD    0xf10a57ef
#define MYMATCHVERSION 0x80000007
#define MYBADDR(bptr) (((ULONG)bptr)<<2)

struct ExtDMPRPort 
	{
	UWORD PixAspX; UWORD PixAspY; 
	UWORD GraphMode;
	APTR BitMap1; APTR BitMap2; APTR BitMap3;
	APTR BitMap4; APTR BitMap5; APTR BitMap6;
	APTR BitMap7; APTR BitMap8;
	UWORD BytesPerRow; UWORD XOffset;
	};

#define PPC_COPIES 32 /* printer can print copies */
#define PPC_MARGINS 128 /* printer driver specifies margins */

struct PrinterExtendedDataTP {
	char	*ped_PrinterName;    /* printer name, null terminated */
	VOID	(*ped_Init)();	     /* called after LoadSeg */
	VOID	(*ped_Expunge)();    /* called before UnLoadSeg */
	int	(*ped_Open)();	     /* called at OpenDevice */
	VOID	(*ped_Close)();      /* called at CloseDevice */
	UBYTE	ped_PrinterClass;    /* printer class */
	UBYTE	ped_ColorClass;      /* color class */
	UBYTE	ped_MaxColumns;      /* number of print columns available */
	UBYTE	ped_NumCharSets;     /* number of character sets */
	UWORD	ped_NumRows;	     /* number of 'pins' in print head */
	ULONG	ped_MaxXDots;	     /* number of dots max in a raster dump */
	ULONG	ped_MaxYDots;	     /* number of dots max in a raster dump */
	UWORD	ped_XDotsInch;	     /* horizontal dot density */
	UWORD	ped_YDotsInch;	     /* vertical dot density */
	char	***ped_Commands;     /* printer text command table */
	int	(*ped_DoSpecial)();  /* special command handler */
	int	(*ped_Render)();     /* raster render function */
	LONG	ped_TimeoutSecs;     /* good write timeout */
	/* the following only exists if the segment version is >= 33 */
	char	**ped_8BitChars;     /* conv. strings for the extended font */
	LONG	ped_PrintMode;	     /* set if text printed, otherwise 0 */
	/* the following only exists if the segment version is >= 34 */
	/* ptr to conversion function for all chars */
	int	(*ped_ConvFunc)();
	ULONG	ped_MarginXDots;	     /* left unprintable margin in number of printer dots */
	ULONG	ped_MarginYDots;	     /* top unprintable margin */
};

struct PreferencesTP4 
	{
	struct Preferences p;	/* Die bisherigen Preferences */
	UWORD pd_PWaitEnabled;	/* Liegt in der Device Struktur da... */
	UWORD Brightness;	/* Ab hier eigentlich pd_Stack */
	UWORD Contrast;		/* Brightness, Contrast Wertebereich 50..200 */
	UWORD Gamma;		/* Wertebereich 20..500 */
	UBYTE GfxTop;		/* oberer Rand beim Grafikdruck in 1/10 Zoll */
	UBYTE GfxFF;		/* Formfeed Flag 1 => FF */
	UWORD Preview;		/* to be defined */
	UBYTE Tray;		/* ------------------ */
	UBYTE Quality;
	UBYTE TxtTop;
	UBYTE Poster;		/* fr weitere Parameter */
	UBYTE Medium;
	UBYTE Copies;
	ULONG FontID;
	ULONG FontSizeID;
	ULONG TextColor;
	UBYTE LOffsRest;
	UBYTE TOffsRest;
	UBYTE Pars[10];
	ULONG Sets[10];
	ULONG MenuSeg;
	UBYTE *MatchTable;
	UBYTE *GammaSum;
	UBYTE **GammaTables;
	UWORD SeparationFlags;
	UBYTE TrueMatch;
	UBYTE Smoothing;
	UBYTE Mirror;
	UBYTE Pad;
	ULONG (* colconv24)(ULONG red, ULONG green, ULONG blue);
			/* Farbkonvertierungsroutine R/G/B => CMY */
	APTR  GfxFont;
};


struct tp_printres	/* one of the 7 possible print resolutions */
{
	UWORD dpix, dpiy;
	ULONG max_x, max_y;
	int margins;
	ULONG margin_x, margin_y;
};

enum { BW, GRAY256, COLOR256, RGB16, RGB24 };	/* the color models */

typedef struct gx_device_turboprint
{
	gx_device_printer dev_pri;

	struct tp_printres printres[7];

	int colormodel;	/* as defined above */

	int defaultdensity, density;
	/* controls resolution like the turboprefs / preferences print density */

	int defaultshading, shading;	/* 0=b/W 1=gray 2=color */
	int defaultquality, quality;	/* 0=b/W 1=gray 2=color */
	int defaultsize, size;	
		/* paper size as used in the amiga-os preferences structure */

	int paperlength, printspacing;	
			/* to determine page length; as in WB preferences */

	int defaultcopies, copies, drivercopies; 
		/* drivercopies == 1 if driver handles copies */

	int mirror;
	int bitsperpixel;
	int cl_bitsperpixel, cl_color, cl_density; 
		/* != 0 if specified from command line */

	int cl_quality, cl_zoomx, cl_zoomy, cl_movex, cl_movey;

	struct IORequest *ioreq; /* stuff needed for using the printer.device */
	struct MsgPort *printport;
	struct PrinterData *printDevice;
	struct Preferences *printPrefs;
	struct PrinterExtendedDataTP *printExtData;
	struct PreferencesTP4 *tpPrefs;	
		/* Turboprint specific settings accessible via device structure */

	#ifdef DEBUG
	FILE *debugstream;	/* file handle to the debug window */
	FILE *whooshdebug;
	#endif
} gx_device_turboprint;


/* macro for casting the device argument */

#define xdev ((gx_device_turboprint *) pdev)

/* whoosh macros!: */

#define Pdev	((gx_device_printer *) pdev)
#define Xdev	((gx_device *) pdev)

/********* whoosh: yet another experiment, can be deleted **************/

const gx_device_printer gs_tp24pcx_device =
prn_device(prn_turboprint_procs, "tp24pcx",
	   9 , 9 ,
	   123 , 123 ,	 /*  X_DPI, Y_DPI, */
	   0, 0, 0, 0,		/* margins */
	   24, turboprint_print_page);

/************** end of experiment ***********************/

/********* whoosh: the modern way of doing things follows *******

replaces a lot of the original "wrong" approach,
this new approach will keep Turboprint going for a long long time, 
*/

#define whoosh_get_params gdev_prn_get_params
#define whoosh_put_params gdev_prn_put_params

static const gx_device_procs tp1j_procs =
prn_color_params_procs
	(
	turboprint_open,	/* calls gdev_prn_open at the end */
	gdev_prn_output_page,
	turboprint_close,	/* calls gdev_prn_close at the end */
	gx_default_b_w_map_rgb_color,
	gx_default_b_w_map_color_rgb,
	whoosh_get_params, 
	whoosh_put_params
	) ;

static const gx_device_procs tp8gj_procs =
prn_color_params_procs
	(
	turboprint_open,	/* calls gdev_prn_open at the end */
	gdev_prn_output_page,
	turboprint_close,	/* calls gdev_prn_close at the end */
	gx_default_gray_map_rgb_color,
	gx_default_gray_map_color_rgb,
	whoosh_get_params, 
	whoosh_put_params
	) ;

static const gx_device_procs tp16j_procs =
prn_color_params_procs
	(
	turboprint_open,	/* calls gdev_prn_open at the end */
	gdev_prn_output_page,
	turboprint_close,	/* calls gdev_prn_close at the end */
	turboprint_16_map_rgb_color,
	turboprint_map_color_rgb,
	whoosh_get_params, 
	whoosh_put_params
	);

static const gx_device_procs tp24j_procs =
prn_color_params_procs
	(
	turboprint_open,	/* calls gdev_prn_open at the end */
	gdev_prn_output_page,
	turboprint_close,	/* calls gdev_prn_close at the end */
	gx_default_rgb_map_rgb_color,
	gx_default_rgb_map_color_rgb,
	whoosh_get_params, 
	whoosh_put_params
	);

static const gx_device_procs Tp24j_procs =
prn_color_params_procs
	(
	turboprint_open,	/* calls gdev_prn_open at the end */
	gdev_prn_output_page,
	turboprint_close,	/* calls gdev_prn_close at the end */
	turboprint_24_map_rgb_color,
	turboprint_map_color_rgb,
	whoosh_get_params, 
	whoosh_put_params
	);

#define  whoosh_procs( name_procs , rgb_color , color_rgb )	\
static const gx_device_procs name_procs =\
prn_color_params_procs\
	(\
	turboprint_open,	/* calls gdev_prn_open at the end */\
	gdev_prn_output_page,\
	turboprint_close,	/* calls gdev_prn_close at the end */\
	rgb_color , /* turboprint_24_map_rgb_color, */ \
	color_rgb , /* turboprint_map_color_rgb, */ \
	whoosh_get_params, \
	whoosh_put_params\
	)

#define  tp1j( devname , devstring ) \
const gx_device_turboprint far_data devname \
	=\
	{\
	prn_device_body\
		(\
		gx_device_turboprint, \
		tp1j_procs, \
		devstring ,\
		9 , 9 ,     	/* DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, */\
		123 , 123 ,     /* X_DPI, Y_DPI, */\
		0, 0, 0, 0, \
		1 , 1 , 1 , 0 , 2 , 1 ,\
		turboprint_print_page\
		)\
	}

#define  tp8gj( devname , devstring ) \
const gx_device_turboprint far_data devname \
	=\
	{\
	prn_device_body\
		(\
		gx_device_turboprint, \
		tp8gj_procs, \
		devstring ,\
		9 , 9 ,     	/* DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, */\
		123 , 123 ,     /* X_DPI, Y_DPI, */\
		0, 0, 0, 0, \
		1, 8 , 255 , 0 , 256 , 0 , \
		turboprint_print_page\
		)\
	}

/* pointless: */
#define  tp16j( devname , devstring ) \
const gx_device_turboprint far_data devname \
	=\
	{\
	prn_device_body\
		(\
		gx_device_turboprint, \
		tp16j_procs, \
		devstring ,\
		9 , 9 ,     	/* DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, */\
		123 , 123 ,     /* X_DPI, Y_DPI, */\
		0, 0, 0, 0, \
		3, 16 , 0 , 31 , 1 , 32 , \
		turboprint_print_page\
		)\
	}

#define  tp24j( devname , devstring ) \
const gx_device_turboprint far_data devname \
	=\
	{\
	prn_device_std_body\
		(\
		gx_device_turboprint, \
		tp24j_procs, \
		devstring ,\
		9 , 9 ,     	/* DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, */\
		123 , 123 ,     /* X_DPI, Y_DPI, */\
		0, 0, 0, 0, \
		24, \
		turboprint_print_page\
		)\
	}

#define  Tp24j( devname , devstring ) \
const gx_device_turboprint far_data devname \
	=\
	{\
	prn_device_std_body\
		(\
		gx_device_turboprint, \
		Tp24j_procs, \
		devstring ,\
		9 , 9 ,     	/* DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, */\
		123 , 123 ,     /* X_DPI, Y_DPI, */\
		0, 0, 0, 0, \
		24, \
		turboprint_print_page\
		)\
	}


#ifdef XTURBO
Tp24j( gs_tp24x_device , "tp24x" ) ;
#endif

tp24j( gs_tp24_device , "tp24" ) ;

#ifdef XTURBO
tp24j( gs_Turboprint_device , "Turboprint" ) ;
#endif

/* 
whoosh: 16 bit printing is a waste of time, and has irregular green component,
I dont even think about 15 bit, 32768 colours, makes no sense at all 
when for a small overhead you can have 16777216 colours,
*/
#ifdef XTURBO
tp16j( gs_tp16x_device , "tp16x" ) ;
#endif

tp8gj( gs_tp8g_device , "tp8g" ) ;

tp8gj( gs_tp8_device , "tp8" ) ;

tp1j( gs_tp1_device , "tp1" ) ;

/*************:whoosh **************/

/* page formats as defined in the workbench preferences */

enum { P_US_LETTER, P_US_LEGAL, P_N_TRACTOR, P_W_TRACTOR, P_CUSTOM,
	P_A0, P_A1, P_A2, P_A3, P_A4, P_A5, P_A6, P_A7, P_A8 };

struct PageDefS{
	ULONG width, height;	/* paper size in 1/1000 inch */
	ULONG leftmargin, rightmargin, topmargin, bottommargin;
				/* default margins for printer drivers that don't supply margins */
};

/* amigamar2003: preceded by private, removed because linker probs  */

struct PageDefS pagedef[] = {
	{ 8500, 11000, 200, 300, 125, 500},	/* Letter */
	{ 8500, 14000, 200, 300, 125, 500},	/* Lega l */
	{ 8268, 12000, 125, 143, 125, 0},	/* N.Tractor \        */
	{ 14000,12000, 200, 200, 125, 0},	/* W.Tractor -Length is taken from preferences' paper length */
	{ 8268, 11693, 125, 143, 125, 0},	/* Custom    /        */
	{ 33070,46772, 346, 347, 125, 500},	/* A0 */
	{ 23386,33070, 346, 347, 125, 500},	/* A1 */
	{ 16535,23386, 346, 347, 125, 500},	/* A2 */
	{ 11693,16535, 200, 493, 125, 500},	/* A3 */
	{ 8268, 11693, 125, 143, 125, 500}, 	/* A4 */
	{ 5846, 8268,  125, 143, 125, 500}, 	/* A5 */
	{ 4134, 5846,  125, 143, 125, 500}, 	/* A6 */
	{ 2923, 4134,  134, 134, 125, 500}, 	/* A7 */
	{ 2067, 2923,  134, 134, 125, 500} 	/* A8 */
};

/* how to get inch/1000ths out of points */
#define POINTS_TO_INCH1000THS(x) ((long)(0.49 + (x) * 1000.0 / 72.0) )
#define INCH1000THS_TO_POINTS(x) ((x) * 72.0 / 1000.0)

/* whoosh thing: */
static int cl_read( char *name , int add , int low , int high , int unrestrict )
{
int ret ;

if( 0==env_exists( name ) )return( 0 ) ;

ret = env_int( name , 0 ) ;

if( unrestrict );
else
	{
	if( ret < low )ret = low ;
	if( ret > high )ret = high ;
	}

ret = ret + add ;

return( ret ) ;
}

/* whoosh777: read in parameters from env variables */
static void cl_whoosh( gx_device *pdev )
{
xdev->cl_density = cl_read( "tpdensity" , 0 , 1 , 7 , 0 ) ;
xdev->cl_quality = cl_read( "tpquality" , 1 , 0 , 9 , 0 ) ;

xdev->cl_zoomx = cl_read( "tpzoomx" , 0 , 1 , 999 , 0 ) ;
xdev->cl_zoomy = cl_read( "tpzoomy" , 0 , 1 , 999 , 0 ) ;

xdev->cl_movex = cl_read( "tpmovex" , 0 , 0 , 0 , 1 ) ;
xdev->cl_movey = cl_read( "tpmovey" , 0 , 0 , 0 , 1 ) ;
}

/* ------ Internal routines ------ */

extern gx_device *TP_DEVICE ;

/* Open the printer, and set the margins. */
static int
turboprint_open(gx_device *pdev)
{
	int dpix, dpiy;
	long printable_x, printable_y, margin_x, margin_y; /* in printer pixels */
	long paper_x, paper_y, margin_r, margin_b; /* in printer pixels */
	float mediax, mediay;
	float margins[4];

/*
TP_DEVICE = pdev ;
*/

whoosh_other_set( "tp_other" );

/* 
SetProtection( WHOOSHES , 0 ) ; DeleteFile( WHOOSHES ) ; 
incorporated into whoosh_other_set() 
*/

deb("open") ;

Deb("\ndname={" ) ; Deb( pdev->dname ) ; Deb( "}\n");

	#ifdef DEBUG
	/* open a debug window */
	if (!xdev->debugstream) xdev->debugstream = fopen("con:0/20/320/200/GS8 turbo info:", "w" );
	if (!xdev->whooshdebug) /* xdev->whooshdebug = fopen(TP_ERRORS, "a+" ); */
		{
		SetProtection( TP_ERRORS , 0 ) ;
		DeleteFile( TP_ERRORS ) ;
		xdev->whooshdebug = fopen( TP_ERRORS , "a+" );
		}

	pdebug("--------TURBOPRINT OPEN----------");
	#endif

	/* the turboprint device provides a print_page_copies procedure */

Pdev->printer_procs.print_page_copies = turboprint_print_page_copies;

	/* choose a turboprint paper format that fits the paper */
	/* format selected in the postscript file best */
	/* if there is no paper size chosen, use the turboprefs setting */

	/* get the possible print resolutions, default color model, paper size */

/* 17aug03: */
cl_whoosh( pdev ) ;

/* whoosh: uncommented this in!*/
/* these are set by turboprint_get_params(), transparently */
#ifdef DEBUG
	intdebug("actual cl_bitsperpixel",xdev->cl_bitsperpixel);
	intdebug("actual cl_color",xdev->cl_color);
	intdebug("actual cl_density",xdev->cl_density);
#endif
	/* whoosh: put in the test */
	if( 0==tp_getres(Pdev) )
		{
		intdebug("that wasnt clever whoosh!",0) ;
		return gs_error_unknownerror ;
		}

	/* select a color model */

/* -dColor=0,1 => cl_color==1,2 !!! */

	{
	/* 
	whooshfix: -dBitsPerPixel and -dColor are ignored and redefined 
	via the device name selected
	*/
	switch( pdev->color_info.depth )
		{
		case 24 :
		
		xdev->cl_bitsperpixel = 24 ;
		xdev->cl_color = 2 ;

		break ;
		
		case 16 :
		
		xdev->cl_bitsperpixel = 16 ;
		xdev->cl_color = 2 ;

		break ;
		
		case 8 :
		
		xdev->cl_bitsperpixel = 8 ;
		if( pdev->color_info.num_components==1 )xdev->cl_color = 1 ;
		else xdev->cl_color = 2 ;

		break ;
		
		case 1 :
		
		xdev->cl_bitsperpixel = 1 ;
		xdev->cl_color = 1 ;

		break ;
		}
	}

/* whoosh: uncommented this in!*/
/* these are set by turboprint_get_params(), transparently */
#ifdef DEBUG
	intdebug("cl_bitsperpixel",xdev->cl_bitsperpixel);
	intdebug("cl_color",xdev->cl_color);
	intdebug("cl_density",xdev->cl_density);
	intdebug("initial bitsperpixel",xdev->bitsperpixel);
	intdebug("initial shading",xdev->shading);
#endif

if (xdev->cl_bitsperpixel)
	{
	/* color model selected via command line -dBitsPerPixel=... */

	xdev->bitsperpixel = xdev->cl_bitsperpixel;

	switch(xdev->bitsperpixel)
		{
		case 1:
			xdev->shading = 0;
			break;
		case 8:
			if (xdev->cl_color == 2 || 
			(xdev->cl_color == 0 && xdev->defaultshading == 2) )
			{
				xdev->shading = 2;
			}
			else
			{
				xdev->shading = 1;
			}
			break;
		case 16:
		case 24:
			xdev->shading = 2;
			break;
		}
	}
else	/* xdev->cl_bitsperpixel==0 */
	{
		/* no BitsPerPixel specified on command line */

		xdev->shading = xdev->defaultshading;

		if (xdev->cl_color == 1) xdev->shading = 0;
		if (xdev->cl_color == 2) xdev->shading = 2;

		if (xdev->shading == 2)
		{
			/* color */
			xdev->bitsperpixel = 24;
		}
		else if (xdev->shading == 1)
		{
			/* grayscale */
			xdev->bitsperpixel = 8;
		}
		else
		{
			/* b&w */
			xdev->bitsperpixel = 1;
		}
	}


#ifdef DEBUG
	intdebug("new bitsperpixel",xdev->bitsperpixel);
	intdebug("new shading",xdev->shading);
#endif

	/* set up ghostscript's parameters according to color model */

switch(xdev->bitsperpixel)
	{
	case 1:
		/* whoosh removed, this is how you dont do it,
        	set_dev_proc(pdev, map_rgb_color, gx_default_b_w_map_rgb_color);
        	set_dev_proc(pdev, map_color_rgb, gx_default_b_w_map_color_rgb);
		*/
		xdev->colormodel = BW;

	break;

	case 8:

	if (xdev->shading != 2)
		{
		/* whoosh removed 
        	set_dev_proc(pdev, map_rgb_color, gx_default_gray_map_rgb_color);
       		set_dev_proc(pdev, map_color_rgb, gx_default_gray_map_color_rgb);
		*/
			xdev->colormodel = GRAY256;
		}
	else
		{
        		/*
			set_dev_proc(pdev, map_rgb_color, pc_8bit_map_rgb_color);
	        	set_dev_proc(pdev, map_color_rgb, pc_8bit_map_color_rgb);
			*/
			xdev->colormodel = COLOR256;
		}
	break;

	case 16:
		/*
        	set_dev_proc(pdev, map_rgb_color, turboprint_16_map_rgb_color);
        	set_dev_proc(pdev, map_color_rgb, turboprint_map_color_rgb);
		*/
		
		xdev->colormodel = RGB16;
		break;
	case 24:
		/* whoosh commented these out 
        	set_dev_proc(pdev, map_rgb_color, turboprint_24_map_rgb_color);
        	set_dev_proc(pdev, map_color_rgb, turboprint_map_color_rgb);
		*/

		xdev->colormodel = RGB24;
		break;
	default:
        	set_dev_proc(pdev, map_rgb_color, gx_default_map_rgb_color);
        	set_dev_proc(pdev, map_color_rgb, gx_default_map_color_rgb);

		xdev->colormodel = BW;
	}


#ifdef DEBUG
	intdebug("new colormodel",xdev->colormodel);
	intdebug(":RGB24==",RGB24);
	intdebug(":RGB16==",RGB16);
	intdebug(":COLOR256==",COLOR256);
	intdebug(":GRAY256==",GRAY256);
	intdebug(":BW==",BW);
#endif


/* whooshfix, remove, really completely and totally bad,
memcpy(&pdev->color_info, &color_info[xdev->colormodel], 
		sizeof(gx_device_color_info));
*/
	/* select quality */

	if (xdev->cl_quality)
	{
		/* quality selected via command line -dQuality=... */

		xdev->quality = xdev->cl_quality - 1;
	}
	else
	{
		xdev->quality = xdev->defaultquality;
	}
	intdebug("Quality",xdev->quality);

	/* select density */

dpix = pdev->x_pixels_per_inch;
dpiy = pdev->y_pixels_per_inch;

if (xdev->cl_density)
	{
	/* density selected via command line -dDensity=... */

	xdev->density = xdev->cl_density - 1;
	}
else if ( dpix == 123 && dpiy == 123)
	{
	/* if none was selected via command line -r..., use turboprefs' setting */

	xdev->density = xdev->defaultdensity;
	}
else
	{
	/* choose the closest available print resolution */
	int i;
	int error;
	int besterror = 9999;

	for (i=0; i<7; i++)
		{
error = abs(dpix - xdev->printres[i].dpix) + abs(dpiy - xdev->printres[i].dpiy);
		if (error < besterror)
			{
			besterror = error;
			xdev->density = i;
			}
		}
	}

gx_device_set_resolution(pdev,
	xdev->printres[xdev->density].dpix,
	xdev->printres[xdev->density].dpiy );

/* check if paper size is still the driver default */

mediax = pdev->MediaSize[0];
mediay = pdev->MediaSize[1];

xdev->size = xdev->defaultsize;

if ( mediax > 58 && mediax < 72 && mediay > 58 && mediay < 72)
	{
	/* set paper size to paper area */

	paper_x = pagedef[xdev->size].width;

/* for some formats we get the page length from the WB PaperLength definition */

	if (xdev->size == P_CUSTOM || xdev->size == P_N_TRACTOR || 
			xdev->size == P_W_TRACTOR)
		{
		paper_y = xdev->paperlength *
				(xdev->printspacing == EIGHT_LPI ? 125 : 167);
		}
	else
		{
		paper_y = pagedef[xdev->size].height;
		}

	paper_x = paper_x * xdev->printres[xdev->density].dpix / 1000;
	paper_y = paper_y * xdev->printres[xdev->density].dpiy / 1000;

	gx_device_set_width_height(pdev, paper_x, paper_y);

	pdebug("paper size set!");
	}
else
	{
	paper_x = pdev->width;
	paper_y = pdev->height;
	}

	/* change the paper size according to zoom factor... */

	/*if (xdev->cl_zoomx)
	{
		paper_x = (50 + paper_x * xdev->cl_zoomx) / 100;
	}
	if (xdev->cl_zoomy)
	{
		paper_y = (50 + paper_y * xdev->cl_zoomy) / 100;
	}

	gx_device_set_width_height(pdev, paper_x, paper_y);
	*/
	/*
	inchdebug("paper width ",POINTS_TO_INCH1000THS( pdev->MediaSize[0] ));
	inchdebug("paper height",POINTS_TO_INCH1000THS( pdev->MediaSize[1] ));
	*/
	inchdebug("paper width ", pdev->MediaSize[0] / 72.0 ) ;
	inchdebug("paper height", pdev->MediaSize[1] / 72.0 ) ;
	inchdebug("width ",(double)pdev->width/pdev->HWResolution[ 0 ] );
	inchdebug("height",(double)pdev->height/pdev->HWResolution[ 1 ] );

	/* change the resolution according to zoom factor... */

if ( xdev->cl_zoomx || xdev->cl_zoomy )
	{
	if (!xdev->cl_zoomx) xdev->cl_zoomx = 100;
	if (!xdev->cl_zoomy) xdev->cl_zoomy = 100;

	if ( xdev->cl_zoomx > 100 && xdev->cl_zoomy > 100 )
		{
		/* we want a section of the page and don't adjust paper size */
		pdev->x_pixels_per_inch	= 
			(50 + pdev->x_pixels_per_inch * xdev->cl_zoomx) / 100;
		pdev->y_pixels_per_inch	= 
			(50 + pdev->y_pixels_per_inch * xdev->cl_zoomy) / 100;

		/* we always use top left as zoom origin; 
		so let's recalculate the offset */

		xdev->cl_movey += 1000 * 
	(pdev->MediaSize[1] / 72.0 - pdev->height / pdev->y_pixels_per_inch );
		}
	else
		{
		/* this is to adjust the paper size as well */
		gx_device_set_resolution(pdev,
			(50 + pdev->x_pixels_per_inch * xdev->cl_zoomx) / 100,
			(50 + pdev->y_pixels_per_inch * xdev->cl_zoomy) / 100 );
		gx_device_set_width_height(pdev, paper_x, paper_y);
		}
	}

	/* Set margins and move the page origin according to left and top margin */

	margin_x = pagedef[xdev->size].leftmargin   * xdev->printres[xdev->density].dpix / 1000;
	margin_y = pagedef[xdev->size].topmargin    * xdev->printres[xdev->density].dpiy / 1000;
	margin_r = pagedef[xdev->size].rightmargin  * xdev->printres[xdev->density].dpix / 1000;
	margin_b = pagedef[xdev->size].bottommargin * xdev->printres[xdev->density].dpiy / 1000;

	if (xdev->printres[xdev->density].margins)
	{
		margin_x = xdev->printres[xdev->density].margin_x;
		margin_y = xdev->printres[xdev->density].margin_y;
	}

	/* if printable width / height available, recompute the right / bottom margin */

	printable_x  = xdev->printres[xdev->density].max_x;
	printable_y  = xdev->printres[xdev->density].max_y;

	if (printable_x)
	{
		margin_r = paper_x - printable_x - margin_x;
		if (margin_r < 0) margin_r = 0;
	}

	if (printable_y)
	{
		margin_b = paper_y - printable_y - margin_y;
		if (margin_b < 0) margin_b = 0;
	}
	else
	{
		xdev->printres[xdev->density].max_y = paper_y - margin_y - margin_b;
	}


	margins[0] = (float)margin_x / pdev->x_pixels_per_inch;
	margins[3] = (float)margin_y / pdev->y_pixels_per_inch;
	margins[2] = (float)margin_r / pdev->x_pixels_per_inch;
	margins[1] = (float)margin_b / pdev->y_pixels_per_inch;

	gx_device_set_margins(pdev, margins, true);

	inchdebug("margin x ", pdev->Margins[0]/pdev->MarginsHWResolution[0]);
	inchdebug("margin y ", pdev->Margins[1]/pdev->MarginsHWResolution[ 1 ] );
	inchdebug("margin x inches ", pdev->HWMargins[0] / 72.0 );
	inchdebug("margin y inches ", pdev->HWMargins[3] / 72.0 );
	inchdebug("margin r inches ", pdev->HWMargins[2] / 72.0 );
	inchdebug("margin b inches ", pdev->HWMargins[1] / 72.0 );

	if ( xdev->cl_movex )
	{
		pdev->Margins[0] += xdev->cl_movex * pdev->MarginsHWResolution[0] / 1000;
	}
	if ( xdev->cl_movey )
	{
		pdev->Margins[1] += xdev->cl_movey * pdev->MarginsHWResolution[1] / 1000;
	}

	if (!tp_openprinterdevice(Pdev))
	{
		return_error(gs_error_invalidfileaccess);
			/* this macro is also a return */
	}
	tp_setprefs(Pdev);	/* do some initialization of the printer.device */
	tp_noprintreq(Pdev);	/* do a noprint to initialize color correction */

	/* general initialization (of "superclass ") */
	{
	int ret ;
	char x[100] ;
	deb("pre");
	ret = gdev_prn_open(pdev);
	sprintf( x , "%d " , ret ) ;
	deb( x ) ;
	deb("post");
	return( ret ) ;
	}
}


/* Close the printer */
static int
turboprint_close(gx_device *pdev)
{
deb("close");

	pdebug("--------TURBOPRINT CLOSE----------");

	tp_closeprinterdevice(Pdev);

	return gdev_prn_close(pdev);

}


static UWORD colormap_bw_table[2] = {0xfff,0x000};


/* Send the page to the printer. */

static int
turboprint_print_page_copies(gx_device_printer *pdev, FILE *prn_stream, int num_copies)
{
	int code = 0;
	int copies, i;

deb("ppc");

	/* num_copies from command line overrides turboprefs copies setting */

	if (num_copies > 1)
	{
		copies = num_copies;
	}
	else
	{
		copies = xdev->defaultcopies;
	}

	if (xdev->drivercopies)
	{
		xdev->copies = copies;
		copies = 1;
	}
	else
	{
		xdev->copies = 1;
	}
	intdebug("copies ", copies);
	intdebug("driver ", xdev->copies);

	for (i=0; code >= 0 && i < copies; i++)
	{
		code = turboprint_print_page(pdev, prn_stream);
	}

	return code;
}

/* whoosh: these were local before,
really they should be done as dynamically allocated local variable pointers,
eg instead of 
static struct ColorMap colormap; 
we should have 

struct ColorMap *colormap ;
as local variable, and then dynamically allocate the storage,
stack space is really non PUBLIC, thus it shouldnt be used for 
OS data structures, although it may not matter here,

but then I have to change the code and make sure 
these local variables always get freed properly,

so the current code is completely correct but not re-entrant, 
*/

static struct ColorMap colormap; 
static struct RastPort rastport;
static struct BitMap bitmap;
static struct ExtDMPRPort ExtStruct;	

int control_on(void) ;
int control_off(void) ;


static void whoosh_free( gx_device_printer *pdev )
{
struct IODRPReq *dumpreq;

dumpreq = (struct IODRPReq *)xdev->ioreq;

/* dumpreq->io_ColorMap = GetColorMap( 256 ) ; */

if(xdev->colormodel==GRAY256)
	{
	FreeColorMap( dumpreq->io_ColorMap ) ;
	dumpreq->io_ColorMap = 0 ;
	}
}

static int
turboprint_print_page(gx_device_printer *pdev, FILE *prn_stream)
{
int happened ;
int code = 0;
/* print page returns the number of printet scan lines ??? or an error < 0 */
/* whoosh: possibly it returns 0 if no error? */
int page_width = pdev->width;
int page_height = pdev->height; 
/*dev_print_scan_lines(pdev);*/	/* number of scan lines */

int max_width = 0;
int max_height = 0;

int top_line = 0;
int band_start = 0;
int band_height = 0;
int band_width = 0;
int trim_height = 0;
int skipped_height = 0;
int last_band = 0;
int dpiy = xdev->printres[xdev->density].dpiy;
byte *in_data_0;
byte *in_data_1;

/* structures for the amiga os turboprint printer.device */

/*
whoosh made global:
struct RastPort rastport;
struct BitMap bitmap;
struct ColorMap colormap; 
struct ExtDMPRPort ExtStruct;	
*/
/* Turboprint specific extension for graphic dumps */

struct IODRPReq *dumpreq;

happened = 0 ;

deb("pp");

	inchdebug("in print width ",(double)pdev->width/pdev->HWResolution[ 0 ] );
	inchdebug("in print height",(double)pdev->height/pdev->HWResolution[ 1 ]);
	/* intdebug("in print scanl.",page_height); ==pdev->height */

	intdebug("dens",xdev->density+1);
	intdebug("dpix",xdev->printres[xdev->density].dpix);
	intdebug("dpiy",xdev->printres[xdev->density].dpiy);

max_width  = xdev->printres[xdev->density].max_x;
max_height = xdev->printres[xdev->density].max_y;

if (max_width && page_width > max_width)
	{
	page_width  = max_width;
	}

if (max_height && page_height > max_height)
	{
	page_height = max_height;
	}

	intdebug("width",page_width);
	intdebug("height",page_height);
	intdebug("max_width",max_width);
	intdebug("max_height",max_height);

	intdebug("bitsperpixel",xdev->bitsperpixel);
	intdebug("shading",xdev->shading);
	intdebug("colormodel",xdev->colormodel);

/* open printer.device */

	/*
	if (!tp_openprinterdevice(pdev))
	{
		return_error(gs_error_invalidfileaccess);
	}
	*/

/* initialize RastPort struct */

InitRastPort(&rastport);
rastport.BitMap = &bitmap;

/* initialize DumpRPort struct */

dumpreq = (struct IODRPReq *)xdev->ioreq;


dumpreq->io_RastPort = &rastport;

/* whoosh: original version: */

/* whoosh function */
control_off() ;


dumpreq->io_ColorMap = &colormap;
dumpreq->io_SrcX = 0;	/* print buffer without offset */
dumpreq->io_SrcY = 0;

ExtStruct.PixAspX = 10;	/* 1:1 Aspect */
ExtStruct.PixAspY = 10;

/* color model specific initialization */

switch(xdev->colormodel)
	{
	case BW:
		/* initialize ColorMap struct */
		
		/* whoosh: original version: */
		
		colormap.Flags = 0;
		colormap.Type = 0;
		colormap.Count = 2;
		colormap.ColorTable = colormap_bw_table;
		
		dumpreq->io_Command = PRD_DUMPRPORT;
		dumpreq->io_Modes = 0;
		
		break;
		
	case COLOR256:
	case GRAY256:
		/* initialize ColorMap struct */

		if( xdev->colormodel==COLOR256 )
			tp_build256colormap(pdev, &colormap);
		else 
			{
			struct ColorMap *cm ;
			unsigned i ;
			UWORD *hi , *lo , H , L ;
			
			cm = GetColorMap( 256 ) ;
			dumpreq->io_ColorMap = cm ; 
			/* SetRGB32CM( ) has a bug! it doesnt set low 4 bits of b*/
			hi = (UWORD *)cm->ColorTable ;
			lo = (UWORD *)cm->LowColorBits ;
			for( i = 0 ; i< 256 ; i++ )
				{
				H = i >> 4 ; L = i & 0xf ; 
				hi[ i ] = (H << 8) + (H<<4) + H ;
				lo[ i ] = (L << 8) + (L<<4) + L ;  
				}
			}

		dumpreq->io_Command = PRD_DUMPRPORT | 0x80;
		dumpreq->io_Modes = (ULONG)&ExtStruct;
		
		ExtStruct.GraphMode = TPFMT_Chunky8 ; /* == 1; *//* 8 bit chunky */
		break;
	
	case RGB16:
		dumpreq->io_Command = PRD_DUMPRPORT | 0x80;
		dumpreq->io_Modes = (ULONG)&ExtStruct;
		/* pdev no good: */
		if( turbostring( pdev->dname ) )
			{
			#ifdef DEB
			PutStr("gm23\n");Flush(Output());
			#endif
			
			ExtStruct.GraphMode = 3 | 32; /* 16Bit CMY direkt */
			}
		else
			{
			#ifdef DEB
			PutStr("gm13\n");Flush(Output());
			#endif
			
			ExtStruct.GraphMode = TPFMT_RGB16 ; /* whoosh version */
			}
		       /* 3 | 32; */ /* 16Bit CMY direkt */
		break;
	
	case RGB24:
		
		dumpreq->io_Command = PRD_DUMPRPORT | 0x80;
		dumpreq->io_Modes = (ULONG)&ExtStruct;
		if( turbostring(pdev->dname ) )
			{
			#ifdef DEB
			PutStr("gm24\n");Flush(Output());
			#endif
			
			ExtStruct.GraphMode = 4 | 32; /* 24Bit CMY direkt */
			}
		else 
			{
			#ifdef DEB
			PutStr("gm14\n");Flush(Output());
			#endif
			
			ExtStruct.GraphMode = TPFMT_RGB24 ; /* whoosh version */
			}
		break;
	}

top_line = 0;	/* the top line of the band we are printing */

/* for each band: send band to printer ... */

do
	{
	myputstr("x1\n");

	inchdebug("top line",(double)top_line/pdev->HWResolution[ 1 ] );

	/* render the current band */

	code = gdev_prn_get_bits(pdev, top_line+1, NULL, &in_data_1);

	myputstr("x2\n");

	code = gdev_prn_get_bits(pdev, top_line,   NULL, &in_data_0);

	myputstr("x3\n");
	
	if ( code < 0 )
		{
		myputstr("path1\n");

		whoosh_free( pdev ) ;
		control_on() ;
		tp_closeprinterdevice(pdev);
		return_error(code);
		};

	myputstr("x4\n");

/* now let's see if the page is organized in bands and how they are organized... */

band_height = (*dev_proc(Xdev, get_band))(Xdev, top_line, &band_start);

	myputstr("x5\n");

	inchdebug("band_height",(double)band_height/pdev->HWResolution[ 1 ] );

	if (band_height == 0)
		{
		band_start = 0;
		band_height = page_height;
		}


	/* is this the last band of the page ? */

	if (band_start + band_height >= page_height)
		{
		last_band = 1;
		band_height = page_height - band_start;
		};

	/* here do some trimming: cut white space on the right side of each band */
	/* cut bottom space on last band */

	myputstr("x6\n");

	band_width = tp_trim_bitmap(pdev, in_data_0, in_data_1 - in_data_0, 
			page_width, band_height, &trim_height);
	if (band_width == 0)  band_width = 1;	/* at least print a bitmap 1 pixel wide */

	inchdebug("band_width ",(double)band_width/pdev->HWResolution[ 0 ] );
	inchdebug("trim_height ",(double)trim_height/pdev->HWResolution[ 1 ]);

	InitBitMap(&bitmap, 1, pdev->width, band_height);
	bitmap.BytesPerRow = in_data_1 - in_data_0;

	intdebug("bytes per row: ",bitmap.BytesPerRow);

	bitmap.Planes[0] = (PLANEPTR) in_data_0;

	dumpreq->io_SrcWidth = band_width;
	dumpreq->io_SrcHeight = trim_height;
	if (dumpreq->io_SrcHeight == 0)  dumpreq->io_SrcHeight = 1;
	dumpreq->io_DestCols = dumpreq->io_SrcWidth;
	dumpreq->io_DestRows = dumpreq->io_SrcHeight;

	myputstr("x7\n");

	tp_setprefs(Pdev);	/* do some initialization of the printer.device */

	myputstr("x8\n");

	happened = control_off() ; /* :whoosh */

	dumpreq->io_Special = (last_band || happened ) ? 0 : SPECIAL_NOFORMFEED;

	if (happened || last_band || trim_height > 0)
		{
		if (trim_height > 0 && skipped_height > 0)
			{
			/* bersprungenen Teil wieder hinzufgen */
			xdev->tpPrefs->GfxTop =    
				((dpiy / 2 + 1000 * skipped_height) / dpiy) / 100;
			xdev->tpPrefs->TOffsRest = 
				((dpiy / 2 + 1000 * skipped_height) / dpiy) % 100;
			}
		else
			{
			xdev->tpPrefs->GfxTop =    0;
			xdev->tpPrefs->TOffsRest = 0;
			}
		
		if( happened )
			{
			dumpreq->io_DestRows = 1 ;
			dumpreq->io_SrcHeight = 1 ;
			}
		/* july2003, was code = DoIO(...); :*/
		code = 0 ;
		if( DoIO(xdev->ioreq) )myputstr("nonzero DoIO()\n");
		skipped_height = band_height - trim_height;
		inchdebug("skipped after print ",(double)skipped_height/
					pdev->HWResolution[ 1 ] );
		}
	else
		{
		code = 0;
		/* Abschnitt berspringen, evtl. am Seitenende weglassen */
		skipped_height += band_height;
		inchdebug("skipped ",(double)skipped_height/pdev->HWResolution[1]);
		}

	myputstr("x9\n");

	if (code)
		{
		myputstr("path2\n");
		whoosh_free( pdev ) ;
		control_on() ;
		tp_closeprinterdevice(pdev);
		return_error(gs_error_ioerror);
		}
	else
		{
		myputstr("path3\n");
		/* code = band_start + band_height;*/ /* is this correct ??? */
				/* almost certainly not: whoosh!*/
		code = 0 ; /* whoosh replacement, redundant actually! */
		}

	myputstr("xa\n");

	top_line = band_start + band_height;

	pdebug("------------");

	myputstr("xb\n");

	} 
	while (last_band == 0 && happened == 0 );

/*tp_closeprinterdevice(pdev);*/

myputstr("xc\n");

whoosh_free( pdev ) ;

myputstr("xd\n");

control_on() ;

return code;
}

/* Turboprint 24 bit RGB -> Color index mapping
   function does a RGB->CMY mapping using the Turboprint printer.device
   there is no Color->RGB mapping, as there is no invers mapping function
*/

#define gx_color_value_to_8bits(cv) ((cv) >> (gx_color_value_bits - 8))

/* amigafeb2003: gx_color_index
turboprint_24_map_rgb_color(gx_device *pdev, gx_color_value r, gx_color_value g,
  gx_color_value b)
*/
gx_color_index turboprint_24_map_rgb_color(gx_device *pdev, const gx_color_value cv[] )
{
/* amigafeb2003: added the next 2 lines */
gx_color_value r , g, b ;

deb("24mrc");

r = cv[ 0 ] ; g = cv[ 1 ] ; b = cv[ 2 ] ;
	return (* xdev->tpPrefs->colconv24)
	   (	gx_color_value_to_8bits(r),
		gx_color_value_to_8bits(g),
		gx_color_value_to_8bits(b) );
}

/* amigafeb2003 gx_color_index
turboprint_16_map_rgb_color(gx_device *pdev, gx_color_value r, gx_color_value g,
  gx_color_value b)
*/
gx_color_index
turboprint_16_map_rgb_color(gx_device *pdev, const gx_color_value cv[] )
{
/* amigafeb2003 put in next 2 lines and {}*/
gx_color_value r, g, b ;

deb("16mrc");

r = cv[ 0 ] ; g = cv[ 1 ] ; b = cv[ 2 ] ;
	{
	gx_color_index index;

	index = (* xdev->tpPrefs->colconv24)
	   (	gx_color_value_to_8bits(r),
		gx_color_value_to_8bits(g),
		gx_color_value_to_8bits(b) );

	return(((index & 0xf80000) >> 8) | ((index & 0xfc00) >> 5) | ((index & 0xf8) >> 3));
	}
}

int
turboprint_map_color_rgb(gx_device *pdev, gx_color_index color,
  gx_color_value prgb[3])
{
deb("mcr");

	pdebug("color_rgb_called");
	return 0;
}

/* Set parameters.  We allow setting:
    BitsPerPixel integer
    Color        integer
    Density      integer  overwritten by resolution -r
*/

static int
turboprint_put_params(gx_device *pdev, gs_param_list *plist)
{
	int v;
	int error = 0;
	int code = 0 ; /* 10july2003 put = 0 */
	int old_bitsperpixel = xdev->cl_bitsperpixel;
	int old_color = xdev->cl_color;
	int old_density = xdev->cl_density;

deb("ppa");

	#ifdef DEBUG
	/* open a debug window */
	if (!xdev->debugstream) xdev->debugstream = fopen("con:0/20/320/200/Debug", "w" );
	if (!xdev->whooshdebug) 
		{
		SetProtection( TP_ERRORS , 0 ) ;
		DeleteFile( TP_ERRORS ) ;
		xdev->whooshdebug = fopen( TP_ERRORS , "a+" );
		}
	#endif
	/* Mark BitsPerPixel as unset */
/* 10jul2003 removed 
	v = 123;

	code = param_read_int(plist, "BitsPerPixel", &v );

myputstr("y1\n");

	if((error == 0) && (code == 0))
	{

myputstr("y2\n");

		if (v != 1 && v != 8 && v != 16 && v != 24)
		{
myputstr("y3\n");
			code = gs_error_rangecheck;
		}
		else
		{
myputstr("y4\n");

			xdev->cl_bitsperpixel = v;
			pdev->color_info.depth = v;	
		*//* to avoid rangecheck in gdev_prn_put_params *//*
		}
	}
	if(code < 0)
	{
		error = code;
myputstr("y5\n");
		param_signal_error(plist,"BitsPerPixel",code);
	}

end of 10july 2003 removal
*/

/* 10jul2003 remove:

	v = 123;

myputstr("y6\n");

	code = param_read_int(plist, "Color", &v );

myputstr("y7\n");

	if((error == 0) && (code == 0))
	{

myputstr("y8\n");

		if (v != 0 && v != 1)
		{
myputstr("y9\n");
			code = gs_error_rangecheck;
		}
		else
		{
myputstr("ya\n");
			xdev->cl_color = v+1;
		}
	}
	if(code < 0)
	{
myputstr("yb\n");
		error = code;
		param_signal_error(plist,"Color",code);
myputstr("yc\n");

	}
end of 10jul2003 removal
*/


	v = 123;

myputstr("yd\n");

	code = param_read_int(plist, "Density", &v );

myputstr("ye\n");

	if((error == 0) && (code == 0))
	{

myputstr("yf\n");
		if (v < 1 || v > 7)
		{
myputstr("zg\n");
			code = gs_error_rangecheck;
		}
		else
		{
myputstr("yh\n");
			xdev->cl_density = v;
		}
	}

myputstr("yi\n");

	if(code < 0)
	{
myputstr("zi1\n");
		error = code;
		param_signal_error(plist,"Density",code);
	}

myputstr("yj\n");

	v = 123;

	code = param_read_int(plist, "Quality", &v );

myputstr("yk\n");

	if((error == 0) && (code == 0))
	{
		if (v < 0 || v > 9)
		{
myputstr("zk1\n");
			code = gs_error_rangecheck;
		}
		else
		{
			xdev->cl_quality = v+1;
		}
	}

myputstr("yl\n");

	if(code < 0)
	{
		error = code;
	
myputstr("zl\n");

		param_signal_error(plist,"Quality",code);
	}

	v = 0;

myputstr("ym\n");

	code = param_read_int(plist, "ZoomX", &v );

myputstr("yn\n");

	if((error == 0) && (code == 0))
	{
		if (v < 1 || v > 999)
		{
myputstr("zn\n");
			code = gs_error_rangecheck;
		}
		else
		{
			xdev->cl_zoomx = v;
		}
	}

myputstr("yo\n");

	if(code < 0)
	{
		error = code;

myputstr("zo\n");
		param_signal_error(plist,"ZoomX",code);
	}

	v = 0;

myputstr("yp\n");

	code = param_read_int(plist, "ZoomY", &v );

myputstr("yq\n");

	if((error == 0) && (code == 0))
	{
myputstr("yq1\n");
		if (v < 1 || v > 999)
		{
myputstr("zq\n");
			code = gs_error_rangecheck;
		}
		else
		{
myputstr("yq3\n");
			xdev->cl_zoomy = v;
		}
	}

myputstr("yr\n");

	if(code < 0)
	{
	myputstr("zr1\n");
		error = code;
		param_signal_error(plist,"ZoomY",code);
	}

myputstr("ys\n");

	v = 0;

	code = param_read_int(plist, "MoveX", &v );

myputstr("yt\n");

	if((error == 0) && (code == 0))
	{
	myputstr("yt1\n");
			xdev->cl_movex = v;
	}
	if(code < 0)
	{

	myputstr("zu\n");

		error = code;
		param_signal_error(plist,"MoveX",code);
	}

	v = 0;

	myputstr("yv\n");

	code = param_read_int(plist, "MoveY", &v );

	myputstr("yw\n");

	if((error == 0) && (code == 0))
	{
			xdev->cl_movey = v;
	}
	if(code < 0)
	{
	myputstr("zx\n");

		error = code;
		param_signal_error(plist,"MoveY",code);
	}

	if ( error < 0 )
	{
	myputstr("zx_\n");
	return error;
	}

	myputstr("yy\n");

	error = gdev_prn_put_params(pdev, plist);

	myputstr("yz\n");

	if ( error < 0 )
	{
	myputstr("zz\n");
		return error;
	}



	if (xdev->cl_bitsperpixel != old_bitsperpixel
		|| xdev->cl_color != old_color
		|| xdev->cl_density != old_density )
	{
		if ( pdev->is_open )
		{
		myputstr("y$\n");

			gs_closedevice(pdev);
		myputstr("y&\n");

		}
	}

	myputstr("y*\n");

	return 0;
}

static int turboprint_get_params(gx_device *pdev, gs_param_list *plist)
{	
int code ;

deb("get_p");

code =  gdev_prn_get_params(pdev, plist);

	if ( code < 0 ||
	    /* 10jul2003 removed:
	(code = param_write_int(plist, "Color", &xdev->cl_color)) < 0 ||
		*/
	    (code = param_write_int(plist, "Density", &xdev->cl_density)) < 0 ||
	    (code = param_write_int(plist, "Quality", &xdev->cl_quality)) < 0 ||
	    (code = param_write_int(plist, "ZoomX", &xdev->cl_zoomx)) < 0 ||
	    (code = param_write_int(plist, "ZoomY", &xdev->cl_zoomy)) < 0 ||
	    (code = param_write_int(plist, "MoveX", &xdev->cl_movex)) < 0 ||
	    (code = param_write_int(plist, "MoveY", &xdev->cl_movey)) < 0
	   )
	  return code;

	return code;
}



/* --------------------------------------- */
/* all internal subroutines start with tp_ */
/* --------------------------------------- */

/* open the printer.device */

static int
tp_trim_bitmap(gx_device_printer *pdev, byte *buffer, int bpr, int width, int height, int *trim_height)
{
	int maxwidth = 0;
	int maxheight = -1;
	int x, y;
	byte *ptr;
	short *ptr16;

deb("trb");

switch(xdev->bitsperpixel)
	{
	case 1:
	for (y = 0; y < height; y++)
		{
		ptr = buffer + y * bpr + (width + 7) / 8;

		for (x = (width + 7) / 8; x; x--)
			{
			if (*(--ptr) != 0)
				{
					break;
				}
			}
			if (x > maxwidth) maxwidth = x;
			if (x != 0) maxheight = y;
		}
		maxwidth = maxwidth * 8;
		break;
	case 8:
		for (y = 0; y < height; y++)
		{
			ptr = buffer + y * bpr + width;

			for (x = width; x; x--)
			{
				if (*(--ptr) != 255)
				{
					break;
				}
			}
			if (x > maxwidth) maxwidth = x;
			if (x != 0) maxheight = y;
		}
		break;
	case 16:
		for (y = 0; y < height; y++)
		{
			ptr16 = (short *)(buffer + y * bpr + 2 * width);

			for (x = width; x; x--)
			{
				if (*(--ptr16) != 0)
				{
					break;
				}
			}
			if (x > maxwidth) maxwidth = x;
			if (x != 0) maxheight = y;
		}
		break;
	case 24:
		for (y = 0; y < height; y++)
		{
			ptr = buffer + y * bpr + 3 * width;

			for (x = width; x; x--)
			{
				if (*(--ptr) != 0 || *(--ptr) != 0 || *(--ptr) != 0)
				{
					break;
				}
			}
			if (x > maxwidth) maxwidth = x;
			if (x != 0) maxheight = y;
		}
		break;
	}

	*trim_height = maxheight + 1;

	if (maxwidth > width) maxwidth = width;

	return maxwidth;
}

static int
tp_openprinterdevice(gx_device_printer *pdev)
{
int ret , opened ;
struct PrinterData *dnode;

ret = 0 ; opened = 0 ;

deb("opprd");

if (xdev->printDevice){ ret = 1 ; goto cleanup; }

if (!(xdev->printport = CreateMsgPort())) 
	{
	tp_faultmessage(pdev,"Create Port failed");
	goto cleanup ;
	}

if (!(xdev->ioreq = (struct IORequest *)
	CreateIORequest(xdev->printport, sizeof(struct IODRPReq)))) 
	{
	/* Disable(); whoosh */

	/* Enable(); */
	tp_faultmessage(pdev,"Create IoRequest failed");
	goto cleanup; /* return(0); */
	}

/* Turboprint geladen? */

dnode = (struct PrinterData *)FindName(&SysBase->DeviceList,"printer.device");

/* load turboprint if not yet present in memory */
if ( (dnode == 0) || ( ((ULONG *)(dnode->pd_OldStk))[2] != MYMATCHWORD ) )
	tp_loadturboprint(pdev);

dnode = (struct PrinterData *)FindName(&SysBase->DeviceList,"printer.device");

{
char *fail ; 
/*
struct IODRPReq *iodrp ;
iodrp = (struct IODRPReq *)dnode ;
*/
struct Device *device ;

fail = 0 ; device = (struct Device *)dnode ;

if( (dnode == 0) )fail = "No printer.device node found\n" ;
else if	(
		(
			(ULONG *)(dnode->pd_OldStk)
		)[2] != MYMATCHWORD 
	)fail = "MYMATCHWORD 0xf10a57ef not matched\n" ;
else if ( device->dd_Library.lib_Version < 39 )
	{
	PutStr("Turboprint Pro 3 or higher required!\n");
	PutStr("however lets see what happens!!!, 10 second delay "
		"before proceeding! type space now to halt!\n");
	Delay( 500 ) ;
	PutStr("OK, here goes, wheeeee!!!\n");
	/* fail = "Turboprint Pro 3 or higher required!\n" */
	}
	/*
	(
		(
			(ULONG *)(dnode->pd_OldStk)
		)[3] < MYMATCHVERSION
	)fail = "less than MYMATCHVERSION 0x80000007\n" ;
	*/

if( fail )
	{
	tp_faultmessage(pdev, fail ) ;
	tp_faultmessage(pdev,"Turboprint not found - start TP first");
	goto cleanup ;
	}
}
/*
if 
( (dnode == 0) || 
	(
		(
			(ULONG *)(dnode->pd_OldStk)
		)[2] != MYMATCHWORD 
	)
	|| 
	( 
		(
			(ULONG *)(dnode->pd_OldStk)
		)[3] < MYMATCHVERSION
	)
)	{
	tp_faultmessage(pdev,"Turboprint not found - start TP first");
	goto cleanup ;
	}
*/

/* success => OpenDevice==0 */
if	(
	OpenDevice("oldprinter.device", 0, xdev->ioreq, 0) 
	/* first try to open turboprint directly */
	&&  
	OpenDevice("printer.device", 0, xdev->ioreq, 0)
	) 
	{ /* (if spooler is present), then use the normal name */
	/* nicht da => turboprint auch nicht ladbar?? */
	tp_faultmessage(pdev,"Can't open printer.device");
	goto cleanup ;
	}
/*
Forbid();
Disable();
*/

/* :so that if xdev->printDevice is set everything was completed */

xdev->printDevice  = (struct PrinterData *)(xdev->ioreq->io_Device);
xdev->printPrefs = (struct Preferences *)&(xdev->printDevice->pd_Preferences);
xdev->tpPrefs = (struct PreferencesTP4 *)xdev->printPrefs;
xdev->printExtData = (struct PrinterExtendedDataTP *)
	&(xdev->printDevice->pd_SegmentData->ps_PED);

/*
Enable();
Permit();
*/

ret = 1 ;

cleanup:

if( ret )return( 1 ) ;

/* whoosh: all failure must come here so these will be freed */
if ( xdev->ioreq )
	{
	DeleteIORequest( xdev->ioreq ) ;
	xdev->ioreq = 0 ;
	}

if( xdev->printport )
	{
	DeleteMsgPort(xdev->printport);	
	xdev->printport = NULL;
	}

return( 0 );
}

/* close the printer.device */

static int tp_closeprinterdevice(gx_device_printer *pdev)
{
deb("cprd");

if (xdev->printDevice && xdev->ioreq)
	{
		{
		/* whoosh added this: */
		/*
		if( 0==strcmp( pdev->dname , "tp8g" ) )Delay(500);
		*/
		/*
		xdev->ioreq->io_Command = CMD_RESET ;
		DoIO(xdev->ioreq);
		*/
		}
	CloseDevice(xdev->ioreq);	
	xdev->printDevice = NULL;
	}

if (xdev->ioreq)DeleteIORequest(xdev->ioreq);		
xdev->ioreq = NULL;

if (xdev->printport)DeleteMsgPort(xdev->printport);	
xdev->printport = NULL;


return(1);
}

/* set turboprint device settings */

static int
tp_setprefs(gx_device_printer *pdev)
{
deb("stprfs");

	xdev->printPrefs->PrintDensity = xdev->density + 1;
	xdev->printPrefs->PrintShade = xdev->shading;
	/* xdev->printPrefs->PaperSize = xdev->papersize; */

	xdev->tpPrefs->Quality = xdev->quality;
	xdev->tpPrefs->GfxFF = 1;

	xdev->tpPrefs->Copies = xdev->copies;

	switch( env_int( "alias" , 2 ) )
		{
		case 0 :
		xdev->tpPrefs->p.PrintFlags |= ANTI_ALIAS;
		break ;

		
		case 2 :
		xdev->tpPrefs->p.PrintFlags &= ANTI_ALIAS;
		break ;

		case 3 :
		xdev->tpPrefs->p.PrintFlags = ANTI_ALIAS;
		break ;
		}
	xdev->tpPrefs->p.PrintFlags &= ~(CENTER_IMAGE);
	xdev->tpPrefs->p.PrintAspect = ASPECT_HORIZ;
	xdev->tpPrefs->Poster = 0;
	xdev->tpPrefs->Smoothing = 0;


	xdev->tpPrefs->p.PrintXOffset = 0;
	xdev->tpPrefs->LOffsRest =      0;

	xdev->tpPrefs->GfxTop =         0;
	xdev->tpPrefs->TOffsRest =      0;

	/*xdev->tpPrefs->Preview = (xdev->colormodel == RGB16 || xdev->colormodel == RGB24 ) ? 4 : 0;*/
			/* 4 => Keine Farbkorrektur mehr */

	return 1;
}

/* get the possible resolutions */

static int
tp_getres(gx_device_printer *pdev)
{
/*
 * Liefert fr den aktuell eingestellten Druckertreiber
 * fr alle "Density" / "Dichte"-Werte
 * die Auflsung und den maximalen Druckbereich
 * initialisiert die Farbkorrektur
 */
	int i;

deb("gtrs");

if (!tp_openprinterdevice(pdev)) return(0);

	xdev->defaultdensity = xdev->printPrefs->PrintDensity - 1;
	xdev->defaultshading = xdev->printPrefs->PrintShade;
	xdev->defaultsize    = xdev->printPrefs->PaperSize >> 4;
	xdev->paperlength    = xdev->printPrefs->PaperLength;
	xdev->printspacing   = xdev->printPrefs->PrintSpacing;
	xdev->defaultquality = xdev->tpPrefs->Quality;
	xdev->mirror = xdev->tpPrefs->Mirror;
	xdev->defaultcopies = xdev->tpPrefs->Copies;
	xdev->drivercopies = 
		((xdev->printExtData->ped_PrinterClass & PPC_COPIES) != 0);

	for (i=0; i<7; i++) {

		xdev->printPrefs->PrintDensity = i+1;

		tp_noprintreq(pdev);

		xdev->printres[i].dpix = xdev->printExtData->ped_XDotsInch;
		xdev->printres[i].dpiy = xdev->printExtData->ped_YDotsInch;
		xdev->printres[i].max_x = xdev->printExtData->ped_MaxXDots;
		xdev->printres[i].max_y = xdev->printExtData->ped_MaxYDots;
		if (xdev->printExtData->ped_PrinterClass & PPC_MARGINS)
		{
			xdev->printres[i].margins = TRUE;
			xdev->printres[i].margin_x = xdev->printExtData->ped_MarginXDots;
			xdev->printres[i].margin_y = xdev->printExtData->ped_MarginYDots;
		}
		else
		{
			xdev->printres[i].margins = FALSE;
		}

		#ifdef WHOOSH_DEBUG
PutStr( "Density " ) ; myPutNum( i+1 ) ; PutStr(": "); 
myPutNum(xdev->printExtData->ped_XDotsInch ) ; PutStr("x");
myPutNum(xdev->printExtData->ped_YDotsInch ) ; PutStr(" Size: ");
myPutNum(xdev->printExtData->ped_MaxXDots) ; PutStr("x");
myPutNum(xdev->printExtData->ped_MaxYDots) ; PutStr(" Offset:(");
myPutNum(xdev->printExtData->ped_MarginXDots) ; PutStr(",");
myPutNum(xdev->printExtData->ped_MarginYDots) ; PutStr(")\n");
		#endif

		#ifdef DEBUG
fprintf(xdev->whooshdebug, "Density %d: %dx%d Size: %ldx%ld Offset:(%ld,%ld)\n", 
		i+1,
		xdev->printExtData->ped_XDotsInch, 
		xdev->printExtData->ped_YDotsInch,
		xdev->printExtData->ped_MaxXDots, 
		xdev->printExtData->ped_MaxYDots,
		xdev->printExtData->ped_MarginXDots, 
		xdev->printExtData->ped_MarginYDots 
		);
fprintf(xdev->debugstream, "Density %d: %dx%d Size: %ldx%ld Offset:(%ld,%ld)\n", 
		i+1,
		xdev->printExtData->ped_XDotsInch, 
		xdev->printExtData->ped_YDotsInch,
		xdev->printExtData->ped_MaxXDots, 
		xdev->printExtData->ped_MaxYDots,
		xdev->printExtData->ped_MarginXDots, 
		xdev->printExtData->ped_MarginYDots 
		);
		#endif
	}

	tp_closeprinterdevice(pdev);

	return(1);
}

/* do a noprint request to get the resolution and printable size */

static int
tp_noprintreq(gx_device_printer *pdev)
{
	struct RastPort dummyRastPort;
	struct BitMap dummyBitMap;
	struct IODRPReq *dumpreq;

deb("noprq");

	InitRastPort(&dummyRastPort);
	dummyRastPort.BitMap = &dummyBitMap;
	InitBitMap(&dummyBitMap,1,10,10);

	dumpreq = (struct IODRPReq *)xdev->ioreq;

	dumpreq->io_Command = PRD_DUMPRPORT;
	dumpreq->io_RastPort = &dummyRastPort;
	dumpreq->io_Modes = 0;
	dumpreq->io_Special = SPECIAL_NOPRINT;

	dumpreq->io_SrcWidth = 10;
	dumpreq->io_SrcHeight = 10;
	dumpreq->io_DestCols = 10;
	dumpreq->io_DestRows = 10;

	dumpreq->io_SrcX = 0;
	dumpreq->io_SrcY = 0;

	DoIO(xdev->ioreq);

	return(1);
}


/* output a turboprint error message */

static int
tp_faultmessage(gx_device_printer *pdev, char *string)
{
deb("fm");

	#ifdef WHOOSH_DEBUG
	PutStr( string ) ; PutStr("\n");
	#endif

	#ifdef	DEBUG
	fprintf(xdev->debugstream, "%s\n", string);
	fprintf(xdev->whooshdebug, "%s\n", string);
	/*fscanf(xdev->debugstream, "%s", dummy);*/
	#else
	PutStr( string ) ; PutStr("\n");	
	#endif

	return 1;
}

static int
tp_printadr(gx_device_printer *pdev, char *string, char *adr)
{
deb("pradr");

	#ifdef WHOOSH_DEBUG
	PutStr( string ) ; PutStr( " " ); myPutNum( (int)adr ) ; myPutNum("\n");
	#endif

	#ifdef	DEBUG
	fprintf(xdev->debugstream, "%s %p\n", string, adr);
	fprintf(xdev->whooshdebug, "%s %p\n", string, adr);
	#endif
	return 1;
}

static int
tp_printdouble(gx_device_printer *pdev, char *string, double i , char *string2 )
{
deb("pr.dbl");

	#ifdef WHOOSH_DEBUG
	PutStr( string ) ; PutStr(" "); myPutFloat( i , 8 ) ; 
	PutStr( string2 ) ;
	PutStr("\n");
	#endif

	#ifdef DEBUG
	fprintf(xdev->debugstream, "%s %f %s\n", string, i , string2 );
	fprintf(xdev->whooshdebug, "%s %f %s\n", string, i , string2 );
	#endif

	return 1;
}

static int
tp_printint(gx_device_printer *pdev, char *string, int i)
{
deb("pr.int");

	#ifdef WHOOSH_DEBUG
	PutStr( string ) ; PutStr(" "); myPutNum( i ) ; PutStr("\n");
	#endif

	#ifdef DEBUG
	fprintf(xdev->debugstream, "%s %d\n", string, i);
	fprintf(xdev->whooshdebug, "%s %d\n", string, i);
	#endif

	return 1;
}

/* load Turboprint if not yet present in memory */

static int
tp_loadturboprint(gx_device_printer *pdev)
{
BPTR nilhandle;
BOOL error = FALSE;
char CommandString[40];

deb("ltp");

if (nilhandle = Open((STRPTR)"NIL:", (LONG) MODE_OLDFILE))
	{
	#ifdef	WHOOSH_DEBUG
	PutStr("Performing command:\n\"turboprint:turboprefs -q\"\n");
	#endif

	sprintf(CommandString,"turboprint:turboprefs -q");
	error = Execute((STRPTR)CommandString,NULL,nilhandle);
	Close(nilhandle);
	}
return(error);
}

/* whoosh, probably I should change these to dynamically allocated local variables */
/* build an Amiga 256 color or 256 gray ColorMap */

static UWORD hi_bits[256]; 
static UWORD lo_bits[256];

static int
tp_build256colormap(gx_device_printer *pdev, struct ColorMap *colormap)
{
/* whoosh version: */

int i ;
gx_color_value rgb[3];


	/* whoosh removed these: have moved hi_bits and lo_bits to be global
	int i, c;
	gx_color_value rgb[3];

	static UWORD hi_bits[256]; 
	static UWORD lo_bits[256];

	UWORD lo, hi;
	*/

	colormap->Flags = 0;
	colormap->Type = 2;
	colormap->Count = 256;
	colormap->ColorTable = (APTR)hi_bits;
	colormap->LowColorBits = (APTR)lo_bits;

deb("b256cm");

/* whoosh version */
for (i=0; i<256; i++)
	{
	unsigned r , g , b ;
	
	(*dev_proc(Xdev, map_color_rgb))
			(Xdev, (gx_color_index)i, rgb);

	r = rgb[ 0 ] >> (gx_color_value_bits - 8 ) ;

	g = rgb[ 1 ] >> (gx_color_value_bits - 8 ) ;

	b = rgb[ 2 ] >> (gx_color_value_bits - 8 ) ;

	/* SetRGB32CM( colormap , i , r << 24 , g << 24 , b << 24 ) ; */
	hi_bits[ i ] = ((r>>4)<<8) + ((g>>4)<<4) + (b>>4) ;
	lo_bits[ i ] = ((r & 0xf )<< 8) + (( g & 0xf )<<4) + (b & 0xf ) ;
	}

/* whoosh: original version:
for (i=0; i<256; i++)
	{
	lo = hi = 0;
	(*dev_proc(Xdev, map_color_rgb))
			(Xdev, (gx_color_index)i, rgb);

	for ( c = 0; c < 3; c++ )
		{	
		byte b = rgb[c] >> (gx_color_value_bits - 8);
		hi = (hi << 4) | (b >> 4);
		lo = (lo << 4) | (b & 0xf);
		}
		hi_bits[i] = hi;
		lo_bits[i] = lo;
	}
*/	

/* from the original: */

	return 0;
}

